from __future__ import annotations

from collections.abc import Iterator
from typing import Any, Self

from pydantic.fields import Field, FieldInfo


class MarshalField:
    """
    Basic boilerplate class for all pydantic-marshals' models' fields.
    This is an interface, and it requires implementing all methods to function
    and to be used in :py:class:`pydantic_marshals.models.base.MarshalModel`
    """

    def __init__(self, alias: str | None = None) -> None:
        """
        :param alias: same as Field(alias=...), can be None for no alias
        """
        self.alias = alias

    @classmethod
    def convert(cls, *source: Any) -> Self | None:
        """
        Convert something into a field.
        If conversion is not possible, this method should return None
        """
        raise NotImplementedError

    def generate_name(self) -> str:
        """
        Generates the name for the field, used in
        :py:meth:`pydantic_marshals.models.base.MarshalModel.generate_model`
        """
        raise NotImplementedError

    def generate_type(self) -> type:
        """
        Generates the type annotation for the field, used in
        :py:meth:`pydantic_marshals.models.base.MarshalModel.generate_model`
        """
        raise NotImplementedError

    # TODO type with TypedDict? Use _FieldInfoInputs?
    def generate_field_data(self) -> Iterator[tuple[str, Any]]:
        """
        Generates field data (kwargs for :py:func:`pydantic.fields.Field`),
        and returns it in the for of an Iterator,
        compatible with the ``dict[str, Any]`` constructor

        Kwarg names and types can be found in
        :py:class:`pydantic.fields._FieldInfoInputs`
        """
        yield "alias", self.alias

    def generate_field(self) -> tuple[type, FieldInfo]:
        """
        Generates field info for the field, used in
        :py:meth:`pydantic_marshals.models.base.MarshalModel.generate_model`
        """
        return (
            self.generate_type(),
            Field(**dict(self.generate_field_data())),
        )



from collections.abc import Callable

from mypy.mro import calculate_mro
from mypy.nodes import (
    GDEF,
    Block,
    ClassDef,
    MypyFile,
    SymbolTable,
    SymbolTableNode,
    TypeInfo,
)
from mypy.plugin import DynamicClassDefContext, Plugin

# TODO redo and generalize
type_matrix = {"pydantic_marshals.models.sqlalchemy.MappedModel"}
methods = {"extend", "create"}


class MarshalsPlugin(Plugin):
    def get_dynamic_class_hook(
        self,
        fullname: str,
    ) -> Callable[[DynamicClassDefContext], None] | None:
        class_fullname, method_name = fullname.rpartition(".")[::2]
        if (
            fullname == "pydantic_marshals.models.sqlalchemy.MappedModel"
            or class_fullname in type_matrix
            and method_name in methods
        ):
            return add_info_hook
        return None

    def get_additional_deps(self, file: MypyFile) -> list[tuple[int, str, int]]:
        return [(10, "pydantic_marshals.mypy.magic", -1)]


def add_info_hook(ctx: DynamicClassDefContext) -> None:
    class_def = ClassDef(ctx.name, Block([]))
    class_def.fullname = ctx.api.qualified_name(ctx.name)
    info = TypeInfo(SymbolTable(), class_def, ctx.api.cur_mod_id)
    class_def.info = info
    bm = ctx.api.named_type("pydantic_marshals.models.base.MarshalBaseModel")
    pm = ctx.api.named_type("pydantic_marshals.mypy.magic.MappedModelStub")
    info.bases = [bm, pm]
    calculate_mro(info)

    ctx.api.add_symbol_table_node(ctx.name, SymbolTableNode(GDEF, info))
    type_matrix.add(class_def.fullname)


def plugin(_: str) -> type[Plugin]:
    return MarshalsPlugin



from collections.abc import Callable
from typing import Any, Self, get_type_hints

from pydantic_marshals.fields.base import MarshalField
from pydantic_marshals.utils import ModeledType


class PropertyField(MarshalField):
    """
    Implementation of :py:class:`MarshalField` to use with properties
    Can be used directly or with an added type_override
    """

    def __init__(
        self,
        mapped_property: property,
        type_override: type | None = None,
        alias: str | None = None,
    ) -> None:
        super().__init__(alias)
        self.mapped_property = mapped_property
        self.type_override = type_override

    @classmethod
    def convert(cls, mapped: Any = None, type_: Any = None, *_: Any) -> Self | None:
        if isinstance(mapped, property) and (type_ is None or isinstance(type_, type)):
            return cls(mapped, type_)
        return None

    @property
    def getter(self) -> Callable[[Any], Any]:
        if self.mapped_property.fget is None:
            raise RuntimeError("Property's fget is None somehow")
        return self.mapped_property.fget

    def generate_name(self) -> str:
        return self.getter.__name__

    def generate_type(self) -> type:
        if self.type_override is not None:
            return self.type_override
        return get_type_hints(self.getter).get("return", Any)


PropertyType = (
    property
    | ModeledType[property]
    | Callable[[Any], Any]
    | ModeledType[Callable[[Any], Any]]
    | PropertyField
)



from typing import Any, TypeVar

from pydantic import BaseModel

T = TypeVar("T")
ModeledType = tuple[T, type[BaseModel]]


def is_subtype(maybe_type: Any, klass: type) -> bool:
    return isinstance(maybe_type, type) and issubclass(maybe_type, klass)



from __future__ import annotations

from collections.abc import Sequence
from typing import Self

from pydantic import BaseModel

from pydantic_marshals.fields.properties import PropertyField, PropertyType
from pydantic_marshals.fields.sqlalchemy.columns import ColumnField, ColumnType
from pydantic_marshals.fields.sqlalchemy.relationships import (
    RelationshipField,
    RelationshipType,
)
from pydantic_marshals.models.base import MarshalModel


class MappedModel(MarshalModel):
    """
    Implementation of :py:class:`MarshalModel` to use inside SQLAlchemy's
    (ORM style) table definitions (i.e. classes inherited from `Base`).

    Right now it can accept and convert into fields:
    - columns (constructed with ``mapped_column()``)
    - relationships (constructed with ``relationship()``)
    - properties (constructed with built-in ``@property()``)
    - properties with overriden type (see :py:class:`PropertyField`)

    An alternative constructor :py:meth:`create` is available to make this happen
    """

    field_types = (
        ColumnField,
        RelationshipField,
        PropertyField,
    )

    @classmethod
    def create(
        cls,
        columns: Sequence[ColumnType] = (),
        relationships: Sequence[RelationshipType] = (),
        properties: Sequence[PropertyType] = (),
        bases: Sequence[type[BaseModel]] = (),
        includes: Sequence[MarshalModel] = (),
    ) -> Self:
        return cls(
            *cls.convert_fields(columns),
            *cls.convert_fields(relationships),
            *cls.convert_fields(properties),
            *(field for model in includes for field in model.fields),
            bases=list(bases),
        )

    def extend(
        self,
        columns: Sequence[ColumnType] = (),
        relationships: Sequence[RelationshipType] = (),
        properties: Sequence[PropertyType] = (),
        bases: Sequence[type[BaseModel]] = (),
        includes: Sequence[MarshalModel] = (),
    ) -> Self:
        return self.create(
            columns=columns,
            relationships=relationships,
            properties=properties,
            bases=bases,
            includes=(self, *includes),
        )



from __future__ import annotations

from typing import Any, Self

from pydantic import BaseModel
from sqlalchemy.orm import Mapped, Relationship

from pydantic_marshals.fields.base import MarshalField
from pydantic_marshals.utils import ModeledType, is_subtype


class RelationshipField(MarshalField):
    """
    Implementation of :py:class:`MarshalField` to use with SQLAlchemy's relationships
    A conversion model has to be specified, preferably one made with
    :py:class:`pydantic_marshals.models.sqlalchemy.MappedModel`
    """

    def __init__(
        self,
        mapped_relationship: Relationship[Any],
        model: type[BaseModel],
        alias: str | None = None,
    ) -> None:
        super().__init__(alias)
        self.relationship = mapped_relationship
        self.model = model

    @classmethod
    def convert(cls, mapped: Any = None, model: Any = None, *_: Any) -> Self | None:
        if isinstance(mapped, Relationship) and is_subtype(model, BaseModel):
            return cls(mapped, model)
        return None

    def generate_name(self) -> str:
        return self.relationship.key

    def generate_type(self) -> type[BaseModel]:
        collection_class = self.relationship.collection_class
        if collection_class is None:
            return self.model
        if not isinstance(collection_class, type):
            raise RuntimeError(f"Collection is not a type: {collection_class}")
        if issubclass(collection_class, list):
            return list[self.model]  # type: ignore[return-value, name-defined]
        raise RuntimeError(f"Bad collection class: {collection_class}")


RelationshipType = (
    ModeledType[Relationship[Any]] | ModeledType[Mapped[Any]] | RelationshipField
)



from __future__ import annotations

from collections.abc import Sequence
from typing import Self

from pydantic import BaseModel

from pydantic_marshals.fields.properties import PropertyType
from pydantic_marshals.fields.sqlalchemy.columns import ColumnType
from pydantic_marshals.fields.sqlalchemy.relationships import RelationshipType
from pydantic_marshals.models.base import MarshalModel

class MappedModelStub:
    @classmethod
    def extend(
        cls,
        columns: Sequence[ColumnType] = (),
        relationships: Sequence[RelationshipType] = (),
        properties: Sequence[PropertyType] = (),
        bases: Sequence[type[BaseModel]] = (),
        includes: Sequence[MarshalModel | MappedModelStub] = (),
    ) -> Self: ...



from collections.abc import Iterator
from typing import Any, ClassVar

from pydantic import BaseModel, create_model

from pydantic_marshals.fields.base import MarshalField


class MarshalBaseModel(BaseModel, from_attributes=True, populate_by_name=True):
    pass


class MarshalModel:
    """
    Basic boilerplate class for all pydantic-marshals models.
    This is a complete class, but it could be extended with alternative constructors,
    updated base model, new runtime methods or auto-converters for fields
    """

    field_types: tuple[type[MarshalField], ...] = ()
    """Field types to be used in :py:meth:`convert_field`"""

    def __init__(
        self,
        *fields: MarshalField,
        bases: list[type[BaseModel]],
    ) -> None:
        """
        :param fields: fields to include in the model
        :param bases: pydantic bases to add to the model
        """
        self.fields: list[MarshalField] = list(fields)
        self.bases: list[type[BaseModel]] = bases
        self._generated_model: type[BaseModel] | None = None

    def __set_name__(self, owner: type, name: str) -> None:
        self.model_name: str = f"{owner.__qualname__}.{name}"

    @classmethod
    def convert_field(cls, raw_field: Any) -> MarshalField:
        """
        Converts "raw_field"s into :py:class:`MarshalField` by trying to call
        the .convert method on all of :py:attr:`field_types` in this class.
        Will raise a `RuntimeError` if none of them return MarshalField
        """
        if isinstance(raw_field, MarshalField):
            return raw_field

        if not isinstance(raw_field, tuple):
            raw_field = (raw_field,)

        for field_type in cls.field_types:
            field: MarshalField | None = field_type.convert(*raw_field)
            if field is not None:
                return field
        raise RuntimeError(f"Couldn't convert field: {raw_field}")

    @classmethod
    def convert_aliased_field(cls, raw_field: Any, alias: str | None) -> MarshalField:
        """
        Same as :py:meth:`.convert_field`, but also adds an ``alias`` to the field
        """
        field = cls.convert_field(raw_field)
        field.alias = alias
        return field

    @classmethod
    def convert_fields(
        cls,
        *fields: Any,
        **aliased_fields: Any,
    ) -> Iterator[MarshalField]:
        """
        Converts all arguments into :py:class:`MarshalField`s
        via :py:meth:`.convert_field` and :py:meth:`.convert_aliased_field`

        :param fields: un-aliased positional arguments for fields, can be tuples
        :param aliased_fields: same as fields, but keys are used as aliases
        :return: yields resulting MappedFields one by one
        :raises RuntimeError: if conversion of one of the fields fails
        """
        yield from (cls.convert_field(field) for field in fields)
        yield from (
            cls.convert_aliased_field(field, alias)
            for alias, field in aliased_fields.items()
        )

    model_base_class: ClassVar[type[BaseModel]] = MarshalBaseModel
    """Base model class. Subclasses of :py:class:`MarshalBaseModel` are recommended"""

    def generate_base(self) -> type[BaseModel] | tuple[type[BaseModel], ...]:
        """Generate __base__ argument for :py:func:`create_model`"""
        return self.model_base_class, *self.bases

    def generate_model(self) -> type[BaseModel]:
        """Generate the pydantic model"""
        return create_model(  # type: ignore[call-overload, no-any-return]
            self.model_name,
            __base__=self.generate_base(),
            **{
                mapped_field.generate_name(): mapped_field.generate_field()
                for mapped_field in self.fields
            },
        )

    @property
    def generated_model(self) -> type[BaseModel]:
        """
        Pydantic model, generated by the MarshalModel
        First call to this will call :py:meth:`.generate_model`,
        then save the result to :py:attr:`._generated_model`,
        for it to be reused on future calls
        """
        if self._generated_model is None:
            self._generated_model = self.generate_model()
        return self._generated_model

    def __get__(
        self,
        instance: Any,  # noqa: U100
        owner: Any | None = None,  # noqa: U100
    ) -> type[BaseModel]:
        return self.generated_model



from __future__ import annotations

from collections.abc import Iterator
from typing import Any, Optional, Self
from warnings import warn

from pydantic_core import PydanticUndefined, PydanticUndefinedType
from sqlalchemy.orm import Mapped, MappedColumn
from sqlalchemy.sql.schema import Column, ScalarElementColumnDefault
from sqlalchemy.sql.sqltypes import String

from pydantic_marshals.fields.base import MarshalField


class ColumnField(MarshalField):
    """
    Implementation of :py:class:`MarshalField` to use with SQLAlchemy's columns
    This will only work with 2.0-style MappedColumns
    """

    def __init__(
        self,
        mapped_column: MappedColumn[Any],
        alias: str | None = None,
    ) -> None:
        super().__init__(alias)
        self.mapped = mapped_column

    @classmethod
    def convert(cls, source: Any = None, *_: Any) -> Self | None:
        if isinstance(source, MappedColumn):
            return cls(source)
        return None

    @property
    def column(self) -> Column[Any]:
        return self.mapped.column

    def generate_name(self) -> str:
        return self.column.name

    def generate_type(self) -> type:
        type_ = self.column.type.python_type
        if self.column.nullable:
            return Optional[type_]  # type: ignore[return-value]  # noqa: NU002
        return type_

    def generate_default(self) -> Any | None | PydanticUndefinedType:
        if self.column.default is None:
            return None if self.column.nullable else PydanticUndefined
        if isinstance(self.column.default, ScalarElementColumnDefault):
            return self.column.default.arg
        warn(f"Default: {self.column.default} is not supported, skipping")
        return PydanticUndefined

    def generate_field_data(self) -> Iterator[tuple[str, Any]]:
        yield from super().generate_field_data()

        yield "default", self.generate_default()

        column_type = self.column.type
        if isinstance(column_type, String):
            yield "max_length", column_type.length


ColumnType = MappedColumn[Any] | Mapped[Any] | ColumnField
